
open Format
open Lexing
open Base
open Ast
open Pretty_print

(* Noms du fichier source *)
let ifile = ref ""

let usage =
  sprintf "usage: %s file.exp" Sys.argv.(0)

(* localise une erreur en indiquant la ligne et la colonne *)
let localisation pos =
  let l = pos.pos_lnum in
  let c = pos.pos_cnum - pos.pos_bol + 1 in
  eprintf "File \"%s\", line %d, characters %d-%d:\n" !ifile l (c-1) c

let localisation2 (pos1,pos2) =
  let l = pos1.pos_lnum in
  let c1 = pos1.pos_cnum - pos1.pos_bol
  and c2 = pos2.pos_cnum - pos1.pos_bol in
  eprintf "File \"%s\", line %d, characters %d-%d:\n" !ifile l c1 c2

let () =
  (* Parsing de la ligne de commande *)
  Arg.parse [] (fun s -> ifile := s) usage ;

  (* On vrifie que le nom du fichier source a bien t indiqu *)
  let ofile = if !ifile="" then begin
    eprintf "Reading on standard input: type in a program...@." ;
    ifile := "/dev/stdin" ;
    "/dev/stdout"
  end else
    (* Ce fichier doit avoir l'extension .exp *)
    if not (Filename.check_suffix !ifile ".exp") then begin
      eprintf "Le fichier d'entre doit avoir l'extension .exp\n@?";
      Arg.usage [] usage;
      exit 1
    end else
      !ifile ^ ".s"
  in
  let ofile = Pervasives.open_out ofile in

  (* Ouverture du fichier source en lecture *)
  let f = open_in !ifile in

  (* Cration d'un tampon d'analyse lexicale *)
  let buf = Lexing.from_channel f in

  try
    (* Parsing: la fonction  Parser.prog transforme le tampon lexical en un
       arbre de syntaxe abstraite si aucune erreur (lexicale ou syntaxique)
       n'est dtecte.
       La fonction Lexer.token est utilise par Parser.prog pour obtenir
       le prochain token. *)
    let p = Parser.prog Lexer.token buf in
    close_in f;

    (* Analyse et traduction vers Tree.program *)
    let p'  = Typing.check_program p in
    let mips = Code.generate_program p' in

      printf "OK:@.@[<2>%a@]@." print_prog p' ;
      fprintf (formatter_of_out_channel ofile)
        "%a@."
        Mips.print_program mips ;
      close_out ofile

  with
    | Lexer.Lexing_error c ->
	(* Erreur lexicale. On rcupre sa position absolue et
	 * on la convertit en numro de ligne *)
	localisation (Lexing.lexeme_start_p buf);
	eprintf "Erreur dans l'analyse lexicale: %c@." c;
	exit 1
    | Typing.TypeError (s,loc)->
	(* Erreur de typage *)
        localisation2 loc;
        eprintf  "Erreur de typage: %s@." s;
        exit 1
    | Parsing.Parse_error ->
	(* Erreur syntaxique. On rcupre sa position absolue et on la
	 * convertit en numro de ligne *)
	localisation (Lexing.lexeme_start_p buf);
	eprintf "Erreur dans l'analyse syntaxique@.";
	exit 1
