
open Base
open Ast

exception TypeError of string * location

module Key = struct type t = string let compare = compare end
module Env = Map.Make(Key)

type profil = asa_typ list * asa_typ

type infoid = Typ of Tree.local * asa_typ | Prof of profil

let tsymb = (Hashtbl.create 17 : (string, string * infoid) Hashtbl.t)

let new_name = 
 let ctr = ref 0 in fun i -> (incr ctr; i^"_"^(string_of_int !ctr))

(* x' is the new name for variable x of type t, it is added to the
   table of symbols - the renaming (as well as the type information) is
   added to the visibility table *)

let add_var x x' l t vis = 
  Hashtbl.add tsymb x' (x,Typ (l,t));
  Env.add x (x',Typ(l,t)) vis
    
let add_fun f f' p vis = 
  Hashtbl.add tsymb f' (f,Prof p);
  Env.add f (f',Prof p) vis

let type_cte = function Int _ -> TypInt | Bool _ -> TypBool

let rec type_expr vis e =
  let reloc x = { asa = x ; loc = e.loc } in
  match e.asa with
  | Cst c -> reloc (Tree.Cst c), type_cte c
  | Var x -> 
      if Env.mem x vis then 
        match Env.find x vis with 
          | (x',Typ (l,t)) -> reloc (Tree.Var (l,x')),t 
          | (_,Prof _) -> raise (TypeError ("function used as ident",e.loc))
      else raise (TypeError ("ident not declared",e.loc))
  | Binop(op,e1,e2) when List.mem op [Plus;Mult;Div;Minus] ->
      let (e1',tau1) = type_expr vis e1 in
      let (e2',tau2) = type_expr vis e2 in
        if tau1 <> TypInt then 
	  raise (TypeError ("Expecting expression of type int",e1.loc))
        else if tau2 <> TypInt then 
	  raise (TypeError ("Expecting expression of type int",e2.loc))
	else reloc (Tree.Binop(op,e1',e2')), TypInt
  | Binop(Eq,e1,e2) ->  
      let (e1',tau1) = type_expr vis e1 
      and (e2',tau2) = type_expr vis e2 in
        if tau1 <> tau2 then 
	  raise (TypeError ("Expecting expression of equal types",e.loc))
	else reloc (Tree.Binop(Eq,e1',e2')), TypBool
  | Binop(rel,e1,e2) ->        
      let (e1',tau1) = type_expr vis e1 
      and (e2',tau2) = type_expr vis e2 in
        if tau1 <> TypInt then 
          raise (TypeError ("Expecting expression of type int",e1.loc))
        else if tau2 <> TypInt then 
          raise (TypeError ("Expecting expression of type int",e2.loc))
        else reloc (Tree.Binop(rel,e1',e2')), TypBool
  | Letin(x,e1,e2)  -> let x' = new_name x in 
      let (e1',tau1) = type_expr vis e1 in
      let vis' = add_var x x' Tree.Local tau1 vis in
      let (e2',tau2) = type_expr vis' e2 in
        reloc (Tree.Letin(x',e1',e2')), tau2
  | If(e1,e2,e3) -> let (e1',tau1) = type_expr vis e1 in
      if tau1<>TypBool then
        raise (TypeError ("Expecting expression of type int",e1.loc))
      else let (e2',tau2) = type_expr vis e2 
      and (e3',tau3) = type_expr vis e3
      in if tau2 <> tau3 then 
        raise (TypeError ("Expecting expression of equal types",e.loc))
      else reloc (Tree.If(e1',e2',e3')), tau2
  | Call(f,le)   ->   
      if Env.mem f vis then 
        match Env.find f vis with 
            (f',Prof(lt,t)) -> 
              let le' = type_args e.loc vis (le,lt) in
                reloc (Tree.Call(f',le')), t
          | (_,Typ _) -> raise (TypeError ("function expected",e.loc))
      else raise (TypeError ("ident not declared",e.loc))

and type_args loc vis = function 
    [],[] -> []
  | e::le,t::lt -> let (e',t') = type_expr vis e in
      if t<>t' then raise (TypeError ("wrong type for argument",e.loc))
      else e'::type_args loc vis (le,lt)
  | _ -> raise (TypeError ("wrong number of arguments",loc))

let rec check_decls vis = function 
  | [] -> []
  | ({asa=Vdecl(id,e)} as d) :: p -> 
      let (e',tau) = type_expr vis e in 
      let id' = new_name id in
      let vis' = add_var id id' Tree.Global tau vis in 
        { d with asa = Tree.Vdecl(id',e') } :: check_decls vis' p 
  | ({asa=Fdecl(id,lv,tr,e)} as d) :: p -> 
      let id' = new_name id in
      let prof = (List.map snd lv,tr) in
      let vis' = add_fun id id' prof vis in
      let lx',vispar = 
        List.fold_right 
          (fun (x,t) (lx,v) -> let x'= new_name x in
             (x'::lx),add_var x x' Tree.Local t v)
          lv ([],vis')
      in 
      let (e',tau') = type_expr vispar e in
        if tau'<>tr then raise (TypeError ("wrong result type",e.loc))
        else { d with asa = Tree.Fdecl(id',lx',e') } :: check_decls vis' p 

let check_program p =  check_decls Env.empty p
